/**
 * Este archivo es parte del proyecto EMA Libre.
 * 
 * Contiene el codigo fuente, parte principal de EMA Libre Carpincho,
 * brindando la funcionalidad de recoleccion de datos
 * de una estacion USB compatible.
 *
 * Este archivo, esta basado en el archivo whid.js del proyectos
 * (https://github.com/haricotbean/WeatherStation/blob/master/whid.js)
 * fue modificado y adaptado a la necesidades del software actual.
 * 
 * La funcion principal de este modulo, es obtener los difrentes datos recolectados
 * por los sensores de una estacion USB, y devolver un vector con los mismos, cada X tiempo,
 * este tiempo es configurado en forma general desde el archivo de configuracion
 * de EMA Libre.
 *
 * LICENSE:  This file is part of EMA Libre.
 * EMA Libre is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * EMA Libre is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with EMA Libre.  If not, see <http://www.gnu.org/licenses/>.
 *
 * @copyright  Copyright (c) 2019 GUGLER (http://www.gugler.com.ar)
 * @license    https://www.gnu.org/licenses/gpl-3.0.html GPL License
 * @version    3.0
 * @link       http://www.gugler.com.ar
 * @since      File available since Release 3.0
 */

const util = require('util'); // Utilidades.
const fs = require('fs'); // Libreria fs para el manejo de archivos.
const C_EMA = require('./../conf/ConfiguracionGeneral.js'); // Configuracion General
const parametros = process.argv; // Se lee el parametro
const SO = process.platform; // Se lee la Sistema Operativo ('aix', 'darwin','freebsd','linux','openbsd','sunos','win32')
var Arquitectura = process.arch; // Se lee la arquitectura ('arm', 'arm64', 'ia32', 'mips', 'mipsel', 'ppc', 'ppc64', 's390', 's390x', 'x32', and 'x64'.)
var ModoConsola = 'gui'; // Modo consola o grafico
if (parametros.indexOf('console') === -1) { ModoConsola = 'gui'; } else { ModoConsola = 'console'; } // Modo EMA
var PlataformaFinal = SO + '_' + Arquitectura + '_' + ModoConsola;
var antes=[];

function RegistrarLog (reg, depurar) {
  if (typeof depurar !== 'undefined') { depurar = false; }
  // se convierte la fecha a la UTC parametrizada
  var d = new Date();
  var utc = d.getTime() + (d.getTimezoneOffset() * 60000);
  var fechaHora = new Date(utc + (3600000 * (C_EMA.ConfGeneral.UTC)));
  var horas = fechaHora.getHours();
  var minutos = fechaHora.getMinutes();
  var segundos = fechaHora.getSeconds();
  var mes = fechaHora.getMonth() + 1;
  var dia = fechaHora.getDate();
  if (dia <= 9) { dia = '0' + dia; }
  if (mes <= 9) { mes = '0' + mes; }
  if (horas < 10) { horas = '0' + horas; }
  if (minutos < 10) { minutos = '0' + minutos; }
  if (segundos < 10) { segundos = '0' + segundos; }
  var fecha = dia + '/' + mes + '/' + fechaHora.getFullYear() + ' ' + horas + ':' + minutos + ':' + segundos;
  var registrologs = fecha + ' - ' + util.format(reg) + '\n';
  if (depurar) {
    // SE DEPURA EL ARCHIVO DE REGISTROS DE LOGS
    fs.writeFile('./logs/debug.log', '', { encoding: 'utf8', autoClose: true }, function (err) {
      if (err) { RegistrarLog('0|ERROR|Depurando archivo registro de logs :' + err); }
    });
  } else {
    // GUARDA LOS REGISTROS DE LOGS
    fs.appendFile('./logs/debug.log', registrologs, { encoding: 'utf8', autoClose: true }, function (err) {
      if (err) { RegistrarLog('0|ERROR|Guardando registro de logs en archivo :' + err); }
    });
  }
  registrologs=null;fecha=null;dia=null;mes=null;segundos=null;minutos=null;horas=null;fechaHora=null;utc=null;d=null;
}

// Se carga el driver USB NODE-HID Segun la plataforma y modalidad


try {
  RegistrarLog('0|CORE|Buscando Modulo USB para el S.O - ' + PlataformaFinal);
  var HID = require('node-hid');
  RegistrarLog("0|CORE|Modulo node-hid cargado correctamente.");
} catch (e) {
  RegistrarLog('0|ERROR|No se pudo levantar el modulo node hid para su plataforma. ' + PlataformaFinal);
}

// CODIGO DEL RECOLECTOR DE DATOS

module.exports = wusb = {

  conf: {
    vid: 0x1941,    // USB Vendedor ID
    pid: 0x8021,    // USB Producto ID
    recent: 10,     // Cantidad de registros para mantener en la memoria.
    plataforma: SO,  // Sistema operativo donde se ejecuta
    debug: 0,       // Debug del recolector
    station: []    // Debug del recolector
  },
  sincronizado: false, // Sincronizado - al primer registro distinto (true).
  recent: [], // Registro reciente
  fecha_registro:'',  // Fecha del ultimo registro
  registro_ant:'', // Registro Anterior
  // Dicccionario de acceso a los datos.
  _map: {
    // rf = Formato de lectura utilizado en buffer circular
    'rf.delay': [0, 'ub'], // Minutos desde la �ltima lectura almacenada
    'rf.hum_in': [1, 'ub'],
    'rf.temp_in': [2, 'ss', 0.1, 1],
    'rf.hum_out': [4, 'ub'],
    'rf.temp_out': [5, 'ss', 0.1, 1],
    'rf.abs_pressure': [7, 'us', 0.1, 1],
    'rf.wind_ave': [9, 'wa', 0.1, 1], // en Metros/Segundo (velocidad media del viento,)
    'rf.wind_gust': [10, 'wg', 0.1, 1], // en Metros/Segundo (r�faga de velocidad del viento)
    'rf.wind_dir': [12, 'ub', 22.5], // Direccion desde el norte
    'rf.rain': [13, 'us', 0.3, 1], // Lluvia Total
    'rf.status': [15, 'bf', ['b1', 'b2', 'b3', 'b4', 'b5', 'lost_sensor_contact', 'rain_overflow', 'b8']],
    'rf.lux': [16, 'luxer', 0.1, 1], // Nuevo registro para luxer.
    'rf.uv': [19, 'ub'], // Nuevo registro para UV.
    // fb = fixed block formats
    'fb.read_period': [16, 'ub'],
    'fb.settings_1': [17, 'bf', ['temp_in_F', 'temp_out_F', 'rain_in',
      'bit3', 'bit4', 'pressure_hPa',
      'pressure_inHg', 'pressure_mmHg'
    ]],
    'fb.settings_2': [18, 'bf', ['wind_mps', 'wind_kmph', 'wind_knot',
      'wind_mph', 'wind_bft', 'bit5',
      'bit6', 'bit7'
    ]],
    'fb.display_1': [19, 'bf', ['pressure_rel', 'wind_gust', 'clock_12hr',
      'date_mdy', 'time_scale_24', 'show_year',
      'show_day_name', 'alarm_time'
    ]],
    'fb.display_2': [20, 'bf', ['temp_out_temp', 'temp_out_chill',
      'temp_out_dew', 'rain_hour', 'rain_day',
      'rain_week', 'rain_month', 'rain_total'
    ]],
    'fb.alarm_1': [21, 'bf', ['bit0', 'time', 'wind_dir', 'bit3',
      'hum_in_lo', 'hum_in_hi',
      'hum_out_lo', 'hum_out_hi'
    ]],
    'fb.alarm_2': [22, 'bf', ['wind_ave', 'wind_gust',
      'rain_hour', 'rain_day',
      'pressure_abs_lo', 'pressure_abs_hi',
      'pressure_rel_lo', 'pressure_rel_hi'
    ]],
    'fb.alarm_3': [23, 'bf', ['temp_in_lo', 'temp_in_hi',
      'temp_out_lo', 'temp_out_hi',
      'wind_chill_lo', 'wind_chill_hi',
      'dew_point_lo', 'dew_point_hi'
    ]],
    'fb.timezone': [24, 'sb'],
    'fb.unknown_01': [25, 'pb'],
    'fb.data_changed': [26, 'ub'],
    'fb.data_count': [27, 'us'],
    'fb.display_3': [29, 'bf', ['illuminance_fc', 'bit1', 'bit2', 'bit3',
      'bit4', 'bit5', 'bit6', 'bit7'
    ]],
    'fb.current_pos': [30, 'us'],
    'fb.rel_pressure': [32, 'us', 0.1, 1],
    'fb.abs_pressure': [34, 'us', 0.1, 1],
    'fb.lux_wm2_coeff': [36, 'us', 0.1],
    'fb.date_time': [43, 'dt'],
    NA: [0, 'XX']
  },
  _pad: function (str, n) {
    return ('                    ' + str).slice(n * -1);
  },
  // Coleccion de utilidades de conversi�n de datos
  _d: {
    // Traduccion de  Formato de offset en block/offset
    'x': function (offset) {
      return [Math.floor(offset / 8), offset % 8];
    },
    'bc': function (data, blk, i) {
      var byte = data[blk][i];
      return ((Math.floor(byte / 16) & 0x0f) * 10) + (byte & 0x0f);
    },
    // Tipo: bit field
    'bf': function (data, blk, i, base, prec) {
      var map = {};
      var lo = data[blk][i];
      var bits = [1, 2, 4, 8, 16, 32, 64, 128];
      for (var i = 0; i < bits.length && i < base.length; i++) {
        map[base[i]] = (lo & bits[i]) > 0;
      }
      return map;
    },
    // Tipo:signed byte
    'sb': function (data, blk, i, base, prec) {
      return (data[blk][i] >= 128) ? (128 - data[blk][i]) : data[blk][i];
    },
    // Tipo: plain byte
    'pb': function (data, blk, i) {
      return data[blk][i];
    },
    // Tipo: unsigned byte
    'ub': function (data, blk, i, base, prec) {
      var res = (data[blk][i] === 0xFF) ? null : data[blk][i];
      if (base) {
        res *= base;
      }
      return (prec) ? Number(res.toFixed(prec)) : res;
    },
    // Tipo: signed short
    'ss': function (data, blk, i, base, prec) {
      var hb = (i + 1 < 8) ? blk : blk + 1;
      var hbi = (i + 1 < 8) ? i + 1 : (7 - i);
      var hi = data[hb][hbi];
      var lo = data[blk][i];
      var res = null;
      if (lo === 0xFF && hi === 0xFF) return res;
      if (hi >= 128) {
        res = ((128 - hi) * 256) - lo;
      } else {
        res = (hi * 256) + lo;
      }
      if (base) {
        res *= base;
      }
      return (prec) ? Number(res.toFixed(prec)) : res;
    },
    // Tipo: unsigned short
    'us': function (data, blk, i, base, prec) {
      var hb = (i + 1 < 8) ? blk : blk + 1;
      var hbi = (i + 1 < 8) ? i + 1 : (7 - i);
      var hi = data[hb][hbi];
      var lo = data[blk][i];
      if (lo === 0xFF && hi === 0xFF) return null;
      var res = (hi * 256) + lo;
      if (base) {
        res *= base;
      }
      return (prec) ? Number(res.toFixed(prec)) : res;
    },
    // formato para lectura de luz:luxer.
    'luxer': function (data, blk, i, base, prec) {
      var hb = (i + 1 < 8) ? blk : blk + 1;
      var hbi = (i + 1 < 8) ? i + 2 : (7 - i);
      var hi = data[hb][hbi];
      var med = data[hb][(hbi - 1)];
      var lo = data[blk][i];
      if (lo === 0xFF && hi === 0xFF) return null;
      var res = (hi * 65536) + (med * 256) + (lo);
      if (base) {
        res *= base;
      }
      return (prec) ? Number(res.toFixed(prec)) : res;
    },

    p: function (v, l) {
      return ('00000' + v).slice(-l);
    },
    'wa': function (data, blk, i, base, prec) {
      var lo = data[blk][i];
      var hi = (data[blk][i + 2] & 0x0f) << 8;
      var res = (hi + lo) * base;
      return (prec) ? Number(res.toFixed(prec)) : res;
    },
    'wg': function (data, blk, i, base, prec) {
      var lo = data[blk][i];
      var hi = (data[blk][i + 1] & 0xf0) << 4;
      var res = (hi + lo) * base;
      return (prec) ? Number(res.toFixed(prec)) : res;
    },
    // Tipo: date time
    'dt': function (data, blk, i, base, prec) {
      var yy = wusb._d.bc(data, blk, i);
      var mm = wusb._d.bc(data, blk, i + 1);
      var dd = wusb._d.bc(data, blk, i + 2);
      var hr = wusb._d.bc(data, blk, i + 3);
      var mi = wusb._d.bc(data, blk, i + 4);
      var ss = wusb._d.bc(data, blk, i + 5);
      return '20' + wusb._d.p(yy, 2) + '-' + wusb._d.p(mm, 2) + '-' + wusb._d.p(dd, 2) + ' ' + wusb._d.p(hr, 2) + ':' + wusb._d.p(mi, 2)+ ':' + wusb._d.p(ss, 2);
    }
  },
  // funcion de conversiones de datos .
  d: function (offset, format, base, prec) {
    if (offset < this.length * 8) {
      var i = wusb._d.x(offset);
      if (wusb._d[format]) {
        return wusb._d[format](this, i[0], i[1], base, prec);
      } else {
        // si no se ha registrado convertidor volcar una representacion de la cadena
        return [offset, '|', i[0], i[1], '>>', this[i[0]][i[1]], '0x' + Number(this[i[0]][i[1]]).toString(16).toUpperCase()].join(',');
      }
    }
  },
  f: function (data, name) {
    var ans = 'NA';
    if (wusb._map[name]) {
      ans = wusb.d.apply(data, wusb._map[name]);
    }
    return ans;               
  },
  decode: function (data, list_or_bool) {
    var block = {};
    if (list_or_bool.length) {
      // ---
    } else {
      // ---
      var pre = (list_or_bool) ? 'fb.' : 'rf.';

      for (var key in wusb._map) {
        if (key.indexOf(pre) === 0) {

          block[key] = wusb.f(data, key);
        }
      }
    }
    return block;


  },
  open: function () {
    try {
      //wusb.conf.plataforma = OS;
      var devices = HID.devices();
      var device = null;
      for (var d in devices) {
        if (devices[d].vendorId === wusb.conf.vid && devices[d].productId === wusb.conf.pid) {
          return new HID.HID(devices[d].path);
        }
      }
    } catch (e) {
      RegistrarLog('2|ERROR|Problema al intentar conectar el USB, desconecte y conecte nuevamente el USB.');
      APP.Configuracion.Errores = "Problema al intentar conectar el USB, desconecte y conecte nuevamente el USB.";
      //throw "Problemas con el USB";
      return false;
    }
    return device;
  },

  status: function () {
  try { 
      var devices = null
      var device = null;
      devices =  HID.devices();
      for (var d in devices) 
      { if (devices[d].vendorId === wusb.conf.vid && devices[d].productId === wusb.conf.pid) { return true; } }
      return false;
    } catch (e) {
      throw "Error al buscar dispositivos con node-hid. "+e;
    }
  },

  // Devuelve un array
  range: function (lo, hi, inc) {
    var v = [];
    var max = (Math.floor(hi / 8) + 1) * 8;
    for (var c = lo; c < max; c += inc) {
      v.push(c);
    }
    return v;
  },
  // Construir un bloque de comandos para enviar a un dispositivo USB
  cmd: function (address) {
    var hi = Math.floor(address / 256);
    var lo = address % 256;
    if (wusb.conf.plataforma === 'win32' || wusb.conf.plataforma === 'win64') { return [0x00, 0xA1, hi, lo, 0x20, 0xA1, hi, lo, 0x20]; } // se le agrega un 0 antes
    else { return [0xA1, hi, lo, 0x20, 0xA1, hi, lo, 0x20]; }
  },
  // Funcion ---
  removeListeners: function (d, f) {
    var listeners = station.listeners(f);
    if (listeners.length > 0) {
      for (var l = 0; l < listeners.length; l++) {
        d.removeListener(f, listeners[l]);
      }
    }
  },
  // Configurar una transferencia masiva y disparar una funcion cuando se completa
  getRange: function (station, to, from, finished) {
    var locs = wusb.range(to, from, 32);
    var count = -1;
    var buf = [];
    // Solicitar el siguiente bloque de datos enviados por el dispositivo USB
    var _next = function () {
      if (locs.length > 0) {
        var addr = locs.shift();
        var cmd = wusb.cmd(addr);
        count = 4; // El n�mero de bytes en la respuesta esperada del dispositivo
        station.write(cmd);
        // RegistrarLog('2|RECOLECTOR|DATOS PUROS: ' + cmd);
      } else {
        // Aseg�rese de que todos los lister hayan sido eliminado.
        wusb.removeListeners(station, 'data');
        finished(buf);
      }
    };
    // Recoger los datos y almacenarlos
    var _chunk = function (data) {
      buf.push(data);
      count--;
      if (count === 0) {
        _next();
      }
    };
    // Aseg�rese de eliminar todos los lister
    wusb.removeListeners(station, 'data');
    station.on('data', _chunk);
    _next();
  },
  diferencia: function(obj1, obj2) {
    const result = {};
    if (Object.is(obj1, obj2)) {
        return undefined;
    }
    if (!obj2 || typeof obj2 !== 'object') {
        return obj2;
    }
    Object.keys(obj1 || {}).concat(Object.keys(obj2 || {})).forEach(key => {
        if(obj2[key] !== obj1[key] && !Object.is(obj1[key], obj2[key])) {
            result[key] = obj2[key];
        }
        if(typeof obj2[key] === 'object' && typeof obj1[key] === 'object') {
            const value = wusb.diferencia(obj1[key], obj2[key]);
            if (value !== undefined) {
                result[key] = value;
            }
        }
    });
    return result;
  },

  lastUpdated: function () {
    var dt = fixed['fb.date_time'] + 'Z' + ((fixed['fb.timezone'] < 0) ? '-' : '+') + fixed['fb.timezone'];
    return dt;
  },
  station: function () {
    return station;
  },
  // consulta a la EMA del ultimo registro.
  Recolectar: function () 
    {
    return new Promise((resultado, error)=>
      {
            var started = process.hrtime();
            wusb.getRange(station, 0x00, 0xff, function (fb) 
            {
              fixed = wusb.decode(fb, true);
              var current = fixed['fb.current_pos'];
              //RegistrarLog('ENTO AL RECOLECTOR');
              wusb.getRange(station, current, current, function (reading) 
                {
                var completed = process.hrtime(started);
                //RegistrarLog('ENTO AL GET');
                //se verifica si es un registro nuevo y los cambios
                if(wusb.registro_ant == ''){wusb.registro_ant = reading;}
                var nuevo =  wusb.decode(reading, false);
                var anterior = wusb.decode(wusb.registro_ant, false);
                var cambios = wusb.diferencia(anterior,nuevo);
                var d = new Date();
                var utc = d.getTime() + (d.getTimezoneOffset() * 60000);
                wusb.fecha_registro =new Date(utc + (3600000 * (C_EMA.ConfGeneral.UTC))); // UTC - Configurado
                delete cambios['rf.delay'];
                if(JSON.stringify(cambios['rf.status']) == "{}") {delete cambios['rf.status'];}
            
                if (JSON.stringify(reading) != JSON.stringify(wusb.registro_ant))
                  {
                    //RegistrarLog('NuevoRegistro:' + wusb.fecha_registro );
                   // RegistrarLog( cambios);
                    wusb.registro_ant =  reading;
                    wusb.sincronizado = true; 
                    resultado([nuevo,fixed,cambios,true]);
                  }else{
                        // carga continua de datos
                        //weather = wusb.decode(reading, false);
                        //wusb.recent.unshift(weather);
                        //wusb.recent = wusb.recent.slice(0, wusb.conf.recent || 5);
                        //weather.unix = (new Date(wusb.lastUpdated())).getTime() - weather['rf.delay'] * 60000;
                        //weather.datetime = new Date(weather.unix);
                        //weather.lastUpdated = new Date(wusb.lastUpdated());  
                        wusb.poller = wusb.Recolectar;
                        resultado([nuevo,fixed,cambios,false]);
                       }
                });
            });
      }).then((resultado)=>{ // Fin del then de la promesa
        //RegistrarLog('ENTO AL THEN DE LA PROMESA');
        var datos = resultado[0];
        var configuraciones = resultado[1];
        // se convierte la informacion recolectada en otras medidas.
        var TemperaturaExternaF = datos['rf.temp_out'] * (9 / 5) + 32;
        var TemperaturaInternaF = datos['rf.temp_in'] * (9 / 5) + 32;
        var VientoEnKm = datos['rf.wind_ave'] * 3.6;
        var RafagaEnKm = datos['rf.wind_gust'] * 3.6;
        var VientoEnMh = datos['rf.wind_ave'] * 2.2369;
        var RafagaEnMh = datos['rf.wind_gust'] * 2.2369;
        var VientoDireccion = '';
        // se convierte la fecha a la UTC parametrizada
        var d = new Date();
        var utc = d.getTime() + (d.getTimezoneOffset() * 60000);
        var db = new Date(utc + (3600000 * (C_EMA.ConfGeneral.UTC))); // UTC - Configurado
        var dbutc = new Date(utc); // UTC Pura
        // Se da formato a la fecha y hora
        var minutos = db.getMinutes();
        if (minutos <= 9) { minutos = '0' + minutos; }
        var hora = db.getHours();
        if (hora <= 9) { hora = '0' + hora; }
        var mes = db.getMonth() + 1;
        if (mes <= 9) { mes = '0' + mes; }
        var dia = db.getDate();
        if (dia <= 9) { dia = '0' + dia; }
        var segundo = db.getSeconds();
        if (segundo <= 9) { segundo = '0' + segundo; }
        var fecha = dia + '/' + mes + '/' + db.getFullYear() + ' ' + hora + ':' + minutos + ':' + segundo;
        var fechaISO = db.getFullYear() + '-' + mes + '-' + dia + 'T' + hora + ':' + minutos + ':' + segundo;
        dia = null;
        mes = null;
        hora = null;
        minutos = null;
        segundo = null;
        db = null;
        utc = null;
        d = null;
        // UTC PURA
        var minutost = dbutc.getMinutes();
        if (minutost <= 9) { minutost = '0' + minutost; }
        var horat = dbutc.getHours();
        if (horat <= 9) { horat = '0' + horat; }
        var mest = dbutc.getMonth() + 1;
        if (mest <= 9) { mest = '0' + mest; }
        var diat = dbutc.getDate();
        if (diat <= 9) { diat = '0' + diat; }
        var segundot = dbutc.getSeconds();
        if (segundot <= 9) { segundot = '0' + segundot; }
        var fechaUTC = dbutc.getFullYear() + '-' + mest + '-' + diat + '+' + horat + ':' + minutost + ':' + segundot;
        diat = null;
        mest = null;
        horat = null;
        minutost = null;
        segundot = null;
        dbutc = null;
        // calculo del punto de rocio
        var punto_de_rocio = (Math.pow((datos['rf.hum_out'] / 100), 0.125)) * (112 - (-(0.9 * datos['rf.temp_out']))) - (-(0.1 * datos['rf.temp_out'])) - 112;
        punto_de_rocio = (Math.round(punto_de_rocio * 100)) / 100;
        var puntoDeRocioF = punto_de_rocio * (9 / 5) + 32;
        var x = datos['rf.wind_dir'];
        // direccion del viento por nombre
        switch (true) {
          case x === 0:
            VientoDireccion = 'Norte';
            break;
          case x > 0 && x < 90:
            VientoDireccion = 'Nordeste';
            break;
          case x === 90:
            VientoDireccion = 'Este';
            break;
          case x > 90 && x < 180:
            VientoDireccion = 'Sudeste';
            break;
          case x === 180:
            VientoDireccion = 'Sur';
            break;
          case x > 180 && x < 270:
            VientoDireccion = 'Sudoeste';
            break;
          case x === 270:
            VientoDireccion = 'Oeste';
            break;
          case x > 270 && x < 360:
            VientoDireccion = 'Noroeste';
            break;
          default:
            VientoDireccion = 'N/A';
        }
        x = null;
        // Se muestra por consola el registro
        if (wusb.conf.debug > 1) RegistrarLog('2|REGISTRO|** Datos recolectados **');
        if (wusb.conf.debug > 1) RegistrarLog('2|REGISTRO|Temperatura Externa: ' + datos['rf.temp_out'] + ' C - ' + TemperaturaExternaF.toFixed(2) + ' F');
        if (wusb.conf.debug > 1) RegistrarLog('2|REGISTRO|Humedad Externa: ' + datos['rf.hum_out'] + '%');
        if (wusb.conf.debug > 1) RegistrarLog('2|REGISTRO|Temperatura Interna: ' + datos['rf.temp_in'] + ' C - ' + TemperaturaInternaF.toFixed(2) + ' F');
        if (wusb.conf.debug > 1) RegistrarLog('2|REGISTRO|Humedad Interna: ' + datos['rf.hum_in'] + '%');
        if (wusb.conf.debug > 1) RegistrarLog('2|REGISTRO|Presion Absoluta: ' + configuraciones['fb.abs_pressure'] + ' hpa');
        if (wusb.conf.debug > 1) RegistrarLog('2|REGISTRO|Presion Relativa: ' + configuraciones['fb.rel_pressure'] + ' hpa');
        if (wusb.conf.debug > 1) RegistrarLog('2|REGISTRO|Viento Direccion: ' + datos['rf.wind_dir'] + ' Direccion: ' + VientoDireccion);
        if (wusb.conf.debug > 1) RegistrarLog('2|REGISTRO|V.Velocidad: ' + datos['rf.wind_ave'] + ' Mps - ' + VientoEnKm.toFixed(2) + ' Km/h - ' + VientoEnMh.toFixed(2) + ' Mph');
        if (wusb.conf.debug > 1) RegistrarLog('2|REGISTRO|V.Rafagas: ' + datos['rf.wind_gust'] + ' Mps - ' + RafagaEnKm.toFixed(2) + ' Km/h - ' + RafagaEnMh.toFixed(2) + ' Mph');
        if (wusb.conf.debug > 1) RegistrarLog('2|REGISTRO|Lluvia Acumulada: ' + datos['rf.rain']);
        if (wusb.conf.debug > 1) RegistrarLog('2|REGISTRO|Punto de Rocio: ' + punto_de_rocio.toFixed(2) + ' C  - ' + puntoDeRocioF.toFixed(2) + ' F');
        // Solo si posee el sensor UV y Lux
        if (datos['rf.lux'] != null && datos['rf.uv'] != null && configuraciones['fb.lux_wm2_coeff'] != null) {
          if (wusb.conf.debug > 1) RegistrarLog('2|REGISTRO|Luz: ' + datos['rf.lux'] + ' lux - ' + (datos['rf.lux'] / 1000).toFixed(2) + ' Klux - ' + (datos['rf.lux'] / 683).toFixed(2) + ' w/m2');
          if (wusb.conf.debug > 1) RegistrarLog('2|REGISTRO|Luz coef: ' + configuraciones['fb.lux_wm2_coeff'].toFixed(2) + ' lux');
          if (wusb.conf.debug > 1) RegistrarLog('2|REGISTRO|UV: ' + datos['rf.uv'] + ' uv');
        }
        // vector para con la informacion recolectada.
        var Vector;
        Vector = [fecha, // 0 Fecha
          fechaISO, // 1 fechaISO
          datos['rf.temp_out'], // 2 Temperatura Ext
          datos['rf.temp_in'], // 3 Temperatura Int
          datos['rf.hum_out'], // 4 Humedad Ext
          datos['rf.hum_in'], // 5 Humedad Int
          datos['rf.lux'], // 6 Luz
          datos['rf.uv'], // 7 UV
          VientoEnKm, // 8 Viento
          RafagaEnKm, // 9 Viento Rafaga
          datos['rf.wind_dir'], // 10 Viento Direc �
          VientoDireccion, // 11 Viento Nombre
          datos['rf.abs_pressure'], // 12 Presion abs
          configuraciones['fb.rel_pressure'], // 13 Presion rel
          datos['rf.rain'], // 14 Luvia total
          punto_de_rocio, // 15 Punto de Rocio
          fechaUTC, // 16 Fecha UTC
          datos['rf.status'] // 17 Estado de los sensores
        ];
        // se borran las variables
        fecha = null;
        fechaISO = null;
        VientoEnKm = null;
        RafagaEnKm = null;
        VientoDireccion = null;
        punto_de_rocio = null;
        fechaUTC = null;
        TemperaturaInternaF = null;
        TemperaturaExternaF = null;
        VientoEnMh = null;
        RafagaEnMh = null;
        puntoDeRocioF = null;
        datos = null;
        configuraciones = null;
        //RegistrarLog('FIN DEL THEN DE LA PROMESA');
        //info return(Vector, cambios, true o false - Si hay cambios o no)
        return [Vector,resultado[2],resultado[3]]; // Finalmente devuelve el vector, con los datos recolectados del USB.
    }); //FIN DEL THEN DE LA PROMESA
  }
};
